/*
 * Copyright (C) 2005 by Robert James Kaes <rjk@wormbytes.ca>
 *   All rights reserved.
 *
 * This code is released into the public domain.  Enjoy.
 *
 * PURPOSE:
 * Generates all possible combinations for a mechanical push button
 * lock like those found in offices and labs in the 1990's.  Locks
 * can have up to eight buttons.  In addition, multi-button pairs
 * are allowed.  For a complete description of what the program is
 * required to accomplish see:
 *   
 * http://www.sentex.net/~mwandel/careers/sample_problem.html
 * 
 */

#include <stdio.h> /* for printf() */
#include <stdlib.h> /* for atoi() */


/* Max number of keys allowed in the combination lock */
#define MAX_KEYS 8


/* Pre-initialize the order for the maximum of eight keys */
static unsigned char g_output[9] = { 1, 2, 3, 4, 5, 6, 7, 8, 0 };

/* Stores the highest key available requested (1 to 8) */
static int g_limit;


/*
 * The pair structure uses two bytes to hold what the pair is.  These
 * values are then converted into bitmasks for multi-pair comparisons
 * and also shifted and merged into a single byte for output.
 */
struct pair_s {
        unsigned char first, second;
};

/* Create the full pair bitmask by joining the two halves */
#define BUILD_MASK(p) ((1 << ((p).first - 1)) | (1 << ((p).second - 1)))

/* Initialize the pair structure with the first possible pair (1 & 2) */
#define INIT_PAIR(p) do { (p).first = 1; (p).second = 2; } while (0)

/*
 * Pairs are stored in the output buffer as bytes encoded with the
 * lower value in the upper nibble and the higher value in the bottom
 * nibble.
 */
#define JOIN_PAIR(p) (((p).first << 4) | (p).second)


/*
 * Initialize the combination array for the combinatoric k-of-n.  See the
 * description for next_choice() below for an explanation of where
 * the function is derived from.
 *
 * Parameters:
 *  v - array containing the combinations generated
 *  n - value of maximum number in set
 *  k - number of values to pull from the set
 */
static void
init_choice_sequence(int *v, int n, int k)
{
        int i;
        for (i = 0; i != k; ++i)
                v[i] = i;
        v[k] = n + 1;
}


/*
 * Choices the next combination of k from the set n.  The algorithm
 * is from "Constructive Combinatorics," Stanton and White.  However,
 * I do not have the book where this algorithm is described.  I found
 * a description of the algorithm online.
 *
 * Parameters:
 *  v - array containing the current last combination
 *  n - value of the max number in the set
 *  k - how many values to pull from the set
 *
 * Returns:
 *  0 - no more choices available
 *  1 - another choice _is_ available
 */
static int
next_choice(int *v, int n, int k)
{
        int i, j;

        if (v[0] >= n - k + 1) return 0;

        j = 0;
        while (v[j + 1] <= v[j] + 1) { ++j; }

        v[j]++;

        for (i = 0; i != j; ++i) v[i] = i;

        return 1;
}


/*
 * Generates the next pair between 1 and max_num.  first and second
 * contain the two halves of the pair.  first is always the lower
 * of the two numbers.
 *
 * Parameters:
 *  pair - pointer to a pair struct.  internally updated by function
 *  max_num - maximum key value allowed
 *
 * Returns:
 *  0 - no more pairs possible
 *  1 - another pair _is_ possible
 */
static int
next_pair(struct pair_s *pair, int max_num)
{
        if (pair->second == max_num) {
                ++pair->first;
                pair->second = pair->first;
        }
        ++pair->second;

        return !(pair->first == max_num);
}


/*
 * Swap the contents of two memory locations.  Uses the XOR swap
 * technique to avoid using another RAM address.
 *
 * NOTE: Using XOR is possible in this application because the
 * values of the keys is limited to integers between 1 and 8.  In
 * general, however, it is better to use a temporary variable during
 * the swap.  For RAM constrained platforms, the XOR solution saves
 * a byte.
 *
 * Parameters:
 *  i - pointer to first value to swap
 *  j - pointer to second value to swap
 */
static inline void
swap_elements(int *i, int *j)
{
        *j ^= *i;
        *i ^= *j;
        *j ^= *i;
}


/*
 * Reverse all elements between first and last; where last points
 * to one past the end of the array (like in the STL.)
 *
 * Parameters:
 *  first - pointer to beginning of sequence
 *  last - one _past_ the end of the sequence
 */
static void
reverse_elements(int *first, int *last)
{
        if (first + 1 == last) return;

        --last;
        while (first < last)
                swap_elements(first++, last--);
}


/*
 * Generates the next permutation of the sequence defined by first..last.
 * Works identically to the next_permutation() function in the C++ STL.
 *
 * Prerequisites:
 *  The sequence first..last must be lexically sorted.
 *
 * Parameters:
 *  first - pointer to beginning of sequence of bytes
 *  last - one _past_ the end of the sequence.
 *
 * Returns:
 *  0 - no more permutations possible
 *  1 - another permutation is possible.
 */
static int
next_permutation(int *first, int *last)
{
        int *i, *j, *k;

        /* Nothing to do for a one-element list */
        if (first + 1 == last) return 0;

        /*
         * i and j are adjacent pointers (with i preceding j.)  The
         * pointers begin at the end of the range and work toward the
         * beginning.
         */
        i = last - 1;
        for (;;) {
                j = i--;
                
                if (*i < *j) {
                        k = last;
                        while (*i > *--k) {}
                        swap_elements(i, k);
                        reverse_elements(j, last);
                        return 1;
                }
                
                if (i == first) {
                        reverse_elements(first, last);
                        return 0;
                }
        }
}


/*
 * Takes a byte and prints either the single number or pair depending
 * on how the byte is encoded.  Pairs are encoded with one number per
 * nibble of the byte.  Declared inline because it's only used by
 * print_permutations() below.
 *
 * Parameters:
 *  p - byte to be decoded and printed
 */
static inline void
print_byte(int p)
{
        if (p & 0xf0)
                printf("%d-", (p >> 4) & 0xf);
        printf("%d ", p & 0xf);
}


/*
 * Displays all the permutations of an array of size bytes.
 *
 * Parameters:
 *  array - pointer to beginning of the array of bytes
 *  size - size of the array.
 */
static void
print_permutations(int *array, int size)
{
        int i;
        
        if (size < 1) return;

        do {
                for (i = 0; i != size; ++i)
                        print_byte(g_output[array[i]]);
                putchar('\n');
        } while (next_permutation(array, array + size));
}


/*
 * Prints out the combinations to standard out.  The keys in the
 * combination are already in the global g_output buffer.  Indexes
 * into this buffer are the values actually rearranged, so that the
 * g_output buffer is not changed between permutation and combination
 * runs.
 */
static void
print_sequences(void)
{
        unsigned char *p;
        int len, num_pairs, i, k;
        int offsets[8], s_offsets[8];

        /*
         * Determine the total length of the buffer, and also how
         * many pairs are present.
         */
        len = num_pairs = 0;
        for (p = g_output; *p; ++p, ++len) {
                offsets[len] = len;
                if (*p & 0xf0) ++num_pairs;
        }

        /*
         * Do the k-of-n choice on the singles and then run the
         * permutations against the pairs already in the buffer.
         */
        for (k = 0; k != len - num_pairs + 1; ++k) {
                init_choice_sequence(s_offsets, len - num_pairs - 1, k);

                do {
                        for (i = 0; i != k; ++i)
                                offsets[num_pairs + i] = s_offsets[i] + num_pairs;
                        print_permutations(offsets, num_pairs + k);
                } while (next_choice(s_offsets, len - num_pairs - 1, k));
        }
}


/*
 * Adds available keys to the output buffer.  Skips any keys
 * already used in any pairs already in the sequence.
 *
 * Parameters:
 *  out - pointer to output buffer for keys
 *  mask - bitmask of keys already used in pairs
 */
static void
gen_remaining_keys(unsigned char *out, unsigned char mask)
{
        int i;

        for (i = 0; i != g_limit; ++i) {
                if ((1 << i) & mask) continue;
                *out++ = i + 1;
        }

        /* terminate the sequence with a NULL sentinel */
        *out = 0;
}


/*
 * A recursive function to generate all the pair sequences.  Each call
 * eliminates two more numbers from the possible pairs available.  The eight
 * key limit means there can be only four simultaneous pairs at one time;
 * therefore, the recursive depth is four.
 *
 * NOTE: The function should be converted to an iterative function if
 * it is ported to a micro with a limited hardware stack (like a
 * Microchip PIC.)
 *
 * Parameters:
 *  out - pointer to the output buffer location for the pair
 *  start - last pair used, so start of list to check
 *  mask - bitmask of pairs already used in the sequence
 *  remain - the number of single keys available for future pairs
 */
static void
make_pair_sequence(unsigned char* out, struct pair_s *start, unsigned char mask, int remain)
{
        struct pair_s pairs;
        unsigned char ourmask;

        /*
         * now iterate all the pairs that do not contain the
         * existing pair.
         */
        pairs = *start;
        do {
                /* skip possibilities that contain our existing pairs */
                ourmask = BUILD_MASK(pairs);
                if (ourmask & mask) continue;

                /* Add to the output */
                *out = JOIN_PAIR(pairs);

                /* create the remaining pairs if there's enough to do so */
                if (remain > 1)
                        make_pair_sequence(out + 1, &pairs, ourmask | mask, remain - 2);
                
                /* Show the output now of the remaining keys */
                gen_remaining_keys(out + 1, ourmask | mask);
                print_sequences();
        } while (next_pair(&pairs, g_limit));
}


/*
 * Driver for the rest of the program.  Accepts the number of keys
 * from the command line and validates the value.
 *
 * Return:
 *  0 - success
 *  1 - usage errors
 */
int
main(int argc, char **argv)
{
        struct pair_s init_pair;
        
        if (--argc < 1) {
                fprintf(stderr, "Usage: lock keys\n");
                return 1;
        }

        g_limit = atoi(argv[1]);
        if (g_limit < 1 || g_limit > MAX_KEYS) {
                fprintf(stderr, "Keys outside of range (1..%d)\n", MAX_KEYS);
                return 1;
        }

        /*
         * The "g_output" sequence is already defined, so set the
         * end-marker and then show the results.
         */
        g_output[g_limit] = 0;
        print_sequences();

        /*
         * Only show the multi-button combinations if there is more
         * than one key.
         */
        if (g_limit < 2) return 0;
        
        INIT_PAIR(init_pair);
        make_pair_sequence(g_output, &init_pair, 0, g_limit);
        return 0;
}
